import networkx as nx
import subprocess
import nodestate
import datetime

def node_visible(nodeobj,time):
    return time > nodeobj.define_time 

def state_at_time(nodeobj,time):
    defined    = nodeobj.define_time 
    submitted  = nodeobj.submit_time
    ready_by   = nodeobj.ready_by_time 
    if time < defined:
        return None
    if defined <= time < submitted:
        return nodestate.DEFINED
    if submitted <= time < ready_by:
        return nodestate.RUNNING
    if ready_by <= time:
        return nodeobj.state

def colorize_graph(dag,normtime = None):
    colorized = nx.DiGraph()
    allnodes = dag.nodes()
    start = min([dag.getNode(n).define_time for n in allnodes])
    stop  = max([dag.getNode(n).ready_by_time for n in allnodes])

    time = start + normtime*(stop-start)

    colorkey = {
        None : None,
        nodestate.DEFINED : 'grey',
        nodestate.RUNNING : 'yellow',
        nodestate.FAILED  : 'red',
        nodestate.SUCCESS : 'green'
    }

    for node in dag.nodes():
        nodeobj = dag.getNode(node)
        
        state   = state_at_time(nodeobj,time)
        color   = colorkey[state]
        visible = node_visible(nodeobj,time)
        
        style = 'filled' if visible else 'invis'
        dot_attr = {'label':'{} '.format(nodeobj.name), 'style':style, 'color': color}

        colorized.add_node(node,dot_attr)
        for pre in dag.predecessors(node):
            colorized.add_edge(pre,node)
    
    dotformat = nx.drawing.nx_pydot.to_pydot(colorized)
    dotformat.set_label(datetime.datetime.fromtimestamp(time).strftime('%Y-%m-%d %H:%M:%S'))
    for e in dotformat.get_edges():
        edge_visible =  node_visible(dag.getNode(e.get_destination().replace('"','')),time)
        if not edge_visible:
            e.set_style('invis')
    return dotformat

def print_dag(dag,name,trackdir,time = None):
    pngfilename = '{}/{}.png'.format(trackdir,name) 
    with open(pngfilename,'w') as pngfile:
        p = subprocess.Popen(['dot','-Tpng','-Gsize=18,12\!','-Gdpi=100'], stdout = pngfile, stdin = subprocess.PIPE)
        p.communicate(colorize_graph(dag,time).to_string())


import logging
import nodestate

log = logging.getLogger(__name__)

def node_ran_and_failed(nodeobj):
    return nodeobj.state == nodestate.FAILED

def upstream_ok(dag,nodeobj):
    upstream = dag.predecessors(nodeobj.identifier)
    log.debug("upstream nodes are %s",dag.predecessors(nodeobj.identifier))
    if not upstream:
        return True
    return all(node_status(dag.getNode(x)) for x in upstream)

def upstream_ready(dag,nodeobj):
    upstream = dag.predecessors(nodeobj.identifier)
    if not upstream:
        return True
    return all(dag.getNode(x).ready() for x in upstream)

def upstream_failure(dag,nodeobj):
    upstream = [dag.getNode(x) for x in dag.predecessors(nodeobj.identifier)]
    if not upstream:
        return False

    log.debug('checking upstream nodes %s',upstream)
    upstream_status = [node_ran_and_failed(obj) or upstream_failure(dag,obj) for obj in upstream]
    log.debug('upstream %s', 'ok' if upstream_status else 'failed')
    return any(upstream_status)

def node_status(nodeobj):
    submitted = nodeobj.submit_time
    ready = nodeobj.ready()
    successful = nodeobj.successful()
    log.debug("node %s: submitted: %s, ready: %s, successful: %s",nodeobj.identifier,submitted,ready,successful)

    return submitted and ready and successful
    
def node_defined_or_waiting(nodeobj):
    running = (nodeobj.state == nodestate.RUNNING)
    defined = (nodeobj.state == nodestate.DEFINED)
    log.debug('defined: %s running %s',defined,running)
    return running or defined



import networkx as nx
import logging
import os
import dagstate
import time
from adage import trackers
from adage import nodestate
from adage.decorators import adageop, adagetask, Rule
from adage.adageobject import adageobject

#silence pyflakes
assert adageop
assert adagetask
assert Rule
assert adageobject

__all__ = ['decorators','trackers']

log = logging.getLogger(__name__)

def validate_finished_dag(dag):
    for node in dag:
        nodeobj = dag.getNode(node)
        if nodeobj.submit_time:
            sanity = all([nodeobj.submit_time > dag.getNode(x).ready_by_time for x in dag.predecessors(node)])
            if not sanity:
                return False
    return True

def nodes_left_or_rule_applicable(adageobj):
    dag,rules = adageobj.dag, adageobj.rules
    nodes_we_could_run = [node for node in dag.nodes() if not dagstate.upstream_failure(dag,dag.getNode(node))]
    nodes_running_or_defined = [x for x in nodes_we_could_run if dagstate.node_defined_or_waiting(dag.getNode(x))]

    if any(rule.applicable(adageobj) for rule in rules):
        return True

    log.debug('nodes we could run: %s',nodes_we_could_run)
    if nodes_running_or_defined:
        log.debug('%s nodes that could be run or are running are left.',len(nodes_running_or_defined))
        log.debug('nodes are: %s',[dag.node[n] for n in nodes_running_or_defined])
        return True
    else:
        log.info('no nodes can be run anymore')
        return False

def update_coroutine(adageobj):
    for i,rule in reversed([x for x in enumerate(adageobj.rules)]):
        if rule.applicable(adageobj):
            do_apply = yield rule
            if do_apply:
                log.info('extending graph.')
                rule.apply(adageobj)
                adageobj.applied_rules.append(adageobj.rules.pop(i))
            yield
        else:
            log.debug('rule not ready yet')

def update_dag(adageobj,decider):
    #iterate rules in reverse so we can safely pop items
    anyapplied = False
    update_loop = update_coroutine(adageobj)
    for possible_rule in update_loop:
        log.info('we could update this with rule: %s',possible_rule)
        command = decider.send((possible_rule,adageobj))
        if command:
            log.debug('we are in fact updating this..')
            anyapplied = True
        update_loop.send(command)
    #we changed the state so let's just recurse
    if anyapplied:
        log.info('we applied a change, so we will recurse to see if we can apply anything else give updated state')
        update_dag(adageobj,decider)

def submit_node(nodeobj,backend):
    nodeobj.resultproxy = backend.submit(nodeobj.task)
    nodeobj.submit_time = time.time()
    if not nodeobj.backend:
        nodeobj.backend = backend

def process_dag(backend,adageobj,decider):
    dag = adageobj.dag
    for node in nx.topological_sort(dag):
        nodeobj = dag.getNode(node)
        log.debug("working on node: %s with obj %s",node,nodeobj)
        if nodeobj.submit_time:
            log.debug("node already submitted. continue")
            continue;
        if dagstate.upstream_ok(dag,nodeobj):
            log.info('Node %s upstream is OK and has not been submitted yet. deciding whether to submit',nodeobj)
            do_submit = decider.send((dag,nodeobj))
            if do_submit:
                log.info('submitting %s job',nodeobj)
                submit_node(nodeobj,backend)
        if dagstate.upstream_failure(dag,nodeobj):
            log.debug('not submitting node: %s due to upstream failure',node)

def update_state(adageobj):
    for node in adageobj.dag.nodes():
        #check node status one last time so we pick up the finishing times
        dagstate.node_status(adageobj.dag.getNode(node))

def trackprogress(trackerlist,adageobj):
    map(lambda t: t.track(adageobj), trackerlist)

def adage_coroutine(backend,extend_decider,submit_decider):
    """the main loop as a coroutine, for sequential stepping"""
    # after priming the coroutin, we yield right away until we get send a state object
    state = yield

    # after receiving the state object, we yield and will start the loop once we regain controls
    yield

    #starting the loop
    while nodes_left_or_rule_applicable(state):
        update_dag(state,extend_decider)
        process_dag(backend,state,submit_decider)
        update_state(state)
        #we're done for this tick, let others proceed
        yield state

def yes_man():
    # we yield until we receive some data via send()
    data = yield
    while True:
        rule, state = data
        log.info('received rule: %s state: %s',rule,state)
        #we yield True and wait again to receive some data
        value = True
        data = yield value

def rundag(adageobj,
           track = False,
           backend = None,
           extend_decider = None,
           submit_decider = None,
           loggername = None,
           workdir = None,
           trackevery = 1,
           update_interval = 0.01,
           additional_trackers = None
           ):

    if loggername:
        global log
        log = logging.getLogger(loggername)

    if not workdir:
        workdir = os.getcwd()

    ## funny behavior of multiprocessing Pools means that
    ## we can not have backendsubmit = multiprocsetup(2)    in the function sig
    ## so we only initialize them here
    if not backend:
        from backends import MultiProcBackend
        backend = MultiProcBackend(2)
    if not extend_decider:
        extend_decider = yes_man()
        extend_decider.next() #prime it..

    if not submit_decider:
        submit_decider = yes_man()
        submit_decider.next() #prime it..

    trackerlist = [trackers.SimpleReportTracker(log,trackevery)]
    if track:
        trackerlist += [trackers.GifTracker(gifname = '{}/workflow.gif'.format(workdir), workdir = '{}/track'.format(workdir))]
        trackerlist += [trackers.TextSnapShotTracker(logfilename = '{}/adagesnap.txt'.format(workdir), mindelta = trackevery)]
        # trackerlist += [trackers.JSONDumpTracker(dumpname = '{}/adage.json'.format(workdir))]

    if additional_trackers:
        trackerlist += additional_trackers

    map(lambda t: t.initialize(adageobj), trackerlist)

    log.info('preparing adage coroutine.')
    coroutine = adage_coroutine(backend,extend_decider,submit_decider)
    coroutine.next() #prime the coroutine....
    coroutine.send(adageobj)


    log.info('starting state loop.')

    try:
        for state in coroutine:
            trackprogress(trackerlist,state)
            time.sleep(update_interval)
    except:
        log.exception('some weird exception caught in adage process loop')
        raise

    if adageobj.rules:
        log.warning('some rules were not applied.')

    log.info('all running jobs are finished.')
    log.info('track last time')

    map(lambda t: t.finalize(adageobj), trackerlist)

    log.info('validating execution')

    if not validate_finished_dag(adageobj.dag):
        log.error('DAG execution not validating')
        raise RuntimeError('DAG execution not validating')
    log.info('execution valid. (in terms of execution order)')

    if any(adageobj.dag.getNode(x).state == nodestate.FAILED for x in adageobj.dag.nodes()):
        log.error('raising RunTimeError due to failed jobs')
        raise RuntimeError('DAG execution failed')



import nodestate
import time
import uuid
import logging

log = logging.getLogger(__name__)

class Node(object):
    def __init__(self,name,task,identifier = None):
        self.identifier = identifier or str(uuid.uuid4())
        self.name = name

        self.task = task

        self.define_time = time.time()
        self.submit_time = None
        self.ready_by_time = None

        self.resultproxy = None
        self.backend = None
        self._state = nodestate.DEFINED

    def __repr__(self):
        return '<Node name: {} id: {} state: {}>'.format(self.name,self.identifier,self.state)

    def update_state(self):
        #if we do not have a result object
        #that means it's not submitted yet
        if not self.resultproxy:
            self._state = nodestate.DEFINED
            return

        #if we have a resultobject
        #but the result is not ready
        #the node is still running
        if not self.ready():
            self._state = nodestate.RUNNING
            return

        #if it's ready it's either successful
        #or failed
        if self.successful():
            self._state = nodestate.SUCCESS
            return
        else:
            self._state = nodestate.FAILED
            return

    @property
    def state(self):
        self.update_state()
        return self._state

    @property
    def result(self):
        return self.backend.result(self.resultproxy)

    def ready(self):
        if not self.resultproxy:
            return False

        ready = self.backend.ready(self.resultproxy)
        if not ready:
            return False

        if not self.ready_by_time:
            self.ready_by_time = time.time()
            log.info('node finished %s finished',self)
        return True

    def successful(self):
        if not self.resultproxy:
            return False
        return self.backend.successful(self.resultproxy)



import adage.graph

class adageobject(object):
    def __init__(self,dag = None, rules = None):
        self.dag = dag or adage.graph.AdageDAG()
        self.rules = rules or []
        self.applied_rules = []
    



import networkx as nx
from node import Node
import logging

log = logging.getLogger(__name__)
class AdageDAG(nx.DiGraph):
    def addTask(self,task, nodename = 'node', depends_on = None):
        node = Node(nodename,task)
        self.addNode(node,depends_on)
        return node
    
    def addNode(self,nodeobj,depends_on = None):
        self.add_node(nodeobj.identifier, {'nodeobj': nodeobj})
        for parent in (depends_on or []):
            self.addEdge(parent,nodeobj)
        
    def addEdge(self,fromobj,toobj):
        self.add_edge(fromobj.identifier,toobj.identifier)

    def getNode(self,ident):
        return self.node[ident]['nodeobj']
        
    def getNodeByName(self,name):
        matching = [x for x in self.nodes() if self.getNode(x).name == name]
        if len(matching) > 1:
            log.error('requested name %s',name)
            log.error('matching nodes %s',[self.getNode(x) for x in matching])
            raise RuntimeError("getting node by name resulted in multiple matching nodes, try getting by ID")
        return self.getNode(matching[0]) if matching else None



import multiprocessing
import sys
import traceback

class MultiProcBackend(object):
    def __init__(self,poolsize):
        self.pool = multiprocessing.Pool(poolsize)

    def submit(self,task):
        return self.pool.apply_async(task)

    def result(self,resultproxy):
        return resultproxy.get()

    def ready(self,resultproxy):
        return resultproxy.ready()

    def successful(self,resultproxy):
        if not self.ready(resultproxy): return False
        return resultproxy.successful()

    def fail_info(self,resultproxy):
        try:
            self.result(resultproxy)
        except:
            t,v,tb =    sys.exc_info()
            traceback.print_tb(tb)
            return (t,v)
            
class CeleryBackend(object):
    def __init__(self,app):
        self.app = app
    def submit(self,task):
        self.app.set_current()
        return task.func.celery.apply_async(task.args,task.kwargs,throw = False)

    def result(self,resultproxy):
        return resultproxy.get()

    def ready(self,resultproxy):
        return resultproxy.ready()

    def successful(self,resultproxy):
        return resultproxy.successful()

    def fail_info(self,resultproxy):
        try:
            self.result(resultproxy)
        except:
            return sys.exc_info()
        
class DummyResultProxy(object):        
    pass
    
class DummyBackend(object):
    def submit(self,task):
        if task:
            pass
        return DummyResultProxy()

    def result(self,resultproxy):
        return None
    
    def ready(self,resultproxy):
        return True
    
    def successful(self,resultproxy):
        return False
    
    def fail_info(self,resultproxy):
        return 'cannot give reason :( '
